"""
Bump versions for all modules on nf-core/modules
or for a single module
"""

import logging
import os
import re
from pathlib import Path
from typing import List, Optional, Tuple, Union

import questionary
import yaml
from rich.box import ROUNDED
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.progress import BarColumn, Progress
from rich.table import Table

import nf_core.modules.modules_utils
import nf_core.utils
from nf_core.components.components_command import ComponentCommand
from nf_core.components.nfcore_component import NFCoreComponent
from nf_core.utils import NFCoreYamlConfig, custom_yaml_dumper, rich_force_colors
from nf_core.utils import plural_s as _s

log = logging.getLogger(__name__)


class ModuleVersionBumper(ComponentCommand):
    def __init__(
        self,
        pipeline_dir: Union[str, Path],
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
    ):
        super().__init__("modules", pipeline_dir, remote_url, branch, no_pull)

        self.up_to_date: List[Tuple[str, str]] = []
        self.updated: List[Tuple[str, str]] = []
        self.failed: List[Tuple[str, str]] = []
        self.ignored: List[Tuple[str, str]] = []
        self.show_up_to_date: Optional[bool] = None
        self.tools_config: Optional[NFCoreYamlConfig]

    def bump_versions(
        self, module: Union[str, None] = None, all_modules: bool = False, show_uptodate: bool = False
    ) -> None:
        """
        Bump the container and conda version of single module or all modules

        Looks for a bioconda tool version in the `main.nf` file of the module and checks whether
        are more recent version is available. If yes, then tries to get docker/singularity
        container links and replace the bioconda version and the container links in the main.nf file
        of the respective module.

        Args:
            module: a specific module to update
            all_modules: whether to bump versions for all modules
        """
        self.up_to_date = []
        self.updated = []
        self.failed = []
        self.ignored = []
        self.show_up_to_date = show_uptodate

        # Check modules directory structure
        self.check_modules_structure()

        # Verify that this is not a pipeline
        if not self.repo_type == "modules":
            raise nf_core.modules.modules_utils.ModuleExceptionError(
                "This command only works on the nf-core/modules repository, not on pipelines!"
            )

        # Get list of all modules
        _, nfcore_modules = nf_core.modules.modules_utils.get_installed_modules(self.directory)

        # Load the .nf-core.yml config
        _, self.tools_config = nf_core.utils.load_tools_config(self.directory)

        # Prompt for module or all
        if module is None and not all_modules:
            question = {
                "type": "list",
                "name": "all_modules",
                "message": "Bump versions for all modules or a single named module?",
                "choices": ["All modules", "Named module"],
            }
            answer = questionary.unsafe_prompt([question], style=nf_core.utils.nfcore_question_style)
            if answer["all_modules"] == "All modules":
                all_modules = True
            else:
                module = questionary.autocomplete(
                    "Tool name:",
                    choices=[m.component_name for m in nfcore_modules],
                    style=nf_core.utils.nfcore_question_style,
                ).unsafe_ask()

        if module:
            self.show_up_to_date = True
            if all_modules:
                raise nf_core.modules.modules_utils.ModuleExceptionError(
                    "You cannot specify a tool and request all tools to be bumped."
                )
            nfcore_modules = [m for m in nfcore_modules if m.component_name == module]
            if len(nfcore_modules) == 0:
                raise nf_core.modules.modules_utils.ModuleExceptionError(
                    f"Could not find the specified module: '{module}'"
                )

        progress_bar = Progress(
            "[bold blue]{task.description}",
            BarColumn(bar_width=None),
            "[magenta]{task.completed} of {task.total}[reset] » [bold yellow]{task.fields[test_name]}",
            transient=True,
            disable=os.environ.get("HIDE_PROGRESS", None) is not None,
        )
        with progress_bar:
            bump_progress = progress_bar.add_task(
                "Bumping nf-core modules versions",
                total=len(nfcore_modules),
                test_name=nfcore_modules[0].component_name,
            )
            for mod in nfcore_modules:
                progress_bar.update(bump_progress, advance=1, test_name=mod.component_name)
                self.bump_module_version(mod)

        self._print_results()

    def bump_module_version(self, module: NFCoreComponent) -> bool:
        """
        Bump the bioconda and container version of a single NFCoreComponent

        Args:
            module: NFCoreComponent
        """
        config_version = None
        bioconda_packages = []
        try:
            # Extract bioconda version from `environment.yml`
            bioconda_packages = self.get_bioconda_version(module)
        except FileNotFoundError:
            # try it in the main.nf instead
            try:
                with open(module.main_nf) as fh:
                    for line in fh:
                        if "bioconda::" in line:
                            bioconda_packages = [b for b in line.split() if "bioconda::" in b]
            except FileNotFoundError:
                log.error(
                    f"Neither `environment.yml` nor `main.nf` of {module.component_name} module could be read to get bioconada version of used tools."
                )

        # If multiple versions - don't update! (can't update mulled containers)
        if not bioconda_packages or len(bioconda_packages) > 1:
            self.failed.append(("Ignoring mulled container", module.component_name))
            return False

        # Don't update if blocked in blacklist
        self.bump_versions_config = getattr(self.tools_config, "bump-versions", {}) or {}
        if module.component_name in self.bump_versions_config:
            config_version = self.bump_versions_config[module.component_name]
            if not config_version:
                self.ignored.append(("Omitting module due to config.", module.component_name))
                return False

        # check for correct version and newer versions
        bioconda_tool_name = bioconda_packages[0].split("=")[0].replace("bioconda::", "").strip("'").strip('"')
        bp = bioconda_packages[0]
        bp = bp.strip("'").strip('"')
        bioconda_version = bp.split("=")[1]

        if not config_version:
            try:
                response = nf_core.utils.anaconda_package(bp)
            except (LookupError, ValueError):
                self.failed.append(
                    (
                        f"Conda version not specified correctly: {Path(module.main_nf).relative_to(self.directory)}",
                        module.component_name,
                    )
                )
                return False

            # Check that required version is available at all
            if bioconda_version not in response.get("versions"):
                self.failed.append((f"Conda package had unknown version: `{module.main_nf}`", module.component_name))
                return False

            # Check version is latest available
            last_ver = response.get("latest_version")
        else:
            last_ver = config_version

        if last_ver is not None and last_ver != bioconda_version:
            log.debug(f"Updating version for {module.component_name}")
            # Get docker and singularity container links
            try:
                docker_img, singularity_img = nf_core.utils.get_biocontainer_tag(bioconda_tool_name, last_ver)
            except LookupError as e:
                self.failed.append((f"Could not download container tags: {e}", module.component_name))
                return False

            patterns = [
                (rf"biocontainers/{bioconda_tool_name}:[^'\"\s]+", docker_img),
                (
                    rf"https://depot.galaxyproject.org/singularity/{bioconda_tool_name}:[^'\"\s]+",
                    singularity_img,
                ),
            ]

            with open(module.main_nf) as fh:
                content = fh.read()

            # Go over file content of main.nf and find replacements
            for pattern in patterns:
                found_match = False
                newcontent = []
                for line in content.splitlines():
                    # Match the pattern
                    matches_pattern = re.findall(rf"^.*{pattern[0]}.*$", line)
                    if matches_pattern:
                        found_match = True

                        # Replace the match
                        newline = re.sub(pattern[0], pattern[1], line)
                        newcontent.append(newline)
                    # No match, keep line as it is
                    else:
                        newcontent.append(line)

                if found_match:
                    content = "\n".join(newcontent) + "\n"
                else:
                    self.failed.append(
                        (f"Did not find pattern {pattern[0]} in module {module.component_name}", module.component_name)
                    )
                    return False

            # Write new content to the file
            with open(module.main_nf, "w") as fh:
                fh.write(content)

            # change version in environment.yml
            if not module.environment_yml:
                log.error(f"Could not read `environment.yml` of {module.component_name} module.")
                return False
            with open(module.environment_yml) as fh:
                env_yml = yaml.safe_load(fh)
            env_yml["dependencies"][0] = re.sub(
                bioconda_packages[0], f"bioconda::{bioconda_tool_name}={last_ver}", env_yml["dependencies"][0]
            )
            with open(module.environment_yml, "w") as fh:
                yaml.dump(env_yml, fh, default_flow_style=False, Dumper=custom_yaml_dumper())

            self.updated.append(
                (
                    f"Module updated:  {bioconda_version} --> {last_ver}",
                    module.component_name,
                )
            )
            return True

        else:
            self.up_to_date.append((f"Module version up to date: {module.component_name}", module.component_name))
            return True

    def get_bioconda_version(self, module: NFCoreComponent) -> List[str]:
        """
        Extract the bioconda version from a module
        """
        # Check whether file exists and load it
        bioconda_packages = []
        if module.environment_yml is not None and module.environment_yml.exists():
            with open(module.environment_yml) as fh:
                env_yml = yaml.safe_load(fh)
            bioconda_packages = env_yml.get("dependencies", [])
        else:
            log.error(f"Could not read `environment.yml` of {module.component_name} module.")

        return bioconda_packages

    def _print_results(self) -> None:
        """
        Print the results for the bump_versions command
        Uses the ``rich`` library to print a set of formatted tables to the command line
        summarising the linting results.
        """

        log.debug("Printing bump_versions results")

        console = Console(force_terminal=rich_force_colors())
        # Find maximum module name length
        max_mod_name_len = 40
        for m in [self.up_to_date, self.updated, self.failed]:
            try:
                max_mod_name_len = max(len(m[2]), max_mod_name_len)
            except Exception:
                pass

        def format_result(module_updates: List[Tuple[str, str]], table: Table) -> Table:
            """
            Create rows for module updates
            """
            # TODO: Row styles don't work current as table-level style overrides.
            # I'd like to make an issue about this on the rich repo so leaving here in case there is a future fix
            last_modname = ""
            row_style = None
            for module_update in module_updates:
                if last_modname and module_update[1] != last_modname:
                    if row_style:
                        row_style = None
                    else:
                        row_style = "magenta"
                last_modname = module_update[1]
                table.add_row(
                    Markdown(f"{module_update[1]}"),
                    Markdown(f"{module_update[0]}"),
                    style=row_style,
                )
            return table

        # Table of up to date modules
        if len(self.up_to_date) > 0 and self.show_up_to_date:
            console.print(
                Panel(
                    rf"[!] {len(self.up_to_date)} Module{_s(self.up_to_date)} version{_s(self.up_to_date)} up to date.",
                    style="bold green",
                )
            )
            table = Table(style="green", box=ROUNDED)
            table.add_column("Module name", width=max_mod_name_len)
            table.add_column("Update Message")
            table = format_result(self.up_to_date, table)
            console.print(table)

        # Table of updated modules
        if len(self.updated) > 0:
            console.print(Panel(rf"[!] {len(self.updated)} Module{_s(self.updated)} updated", style="bold yellow"))
            table = Table(style="yellow", box=ROUNDED)
            table.add_column("Module name", width=max_mod_name_len)
            table.add_column("Update message")
            table = format_result(self.updated, table)
            console.print(table)

        # Table of modules that couldn't be updated
        if len(self.failed) > 0:
            console.print(Panel(rf"[!] {len(self.failed)} Module update{_s(self.failed)} failed", style="bold red"))
            table = Table(style="red", box=ROUNDED)
            table.add_column("Module name", width=max_mod_name_len)
            table.add_column("Update message")
            table = format_result(self.failed, table)
            console.print(table)

        # Table of modules ignored due to `.nf-core.yml`
        if len(self.ignored) > 0:
            console.print(Panel(rf"[!] {len(self.ignored)} Module update{_s(self.ignored)} ignored", style="grey58"))
            table = Table(style="grey58", box=ROUNDED)
            table.add_column("Module name", width=max_mod_name_len)
            table.add_column("Update message")
            table = format_result(self.ignored, table)
            console.print(table)



"""Tests covering the modules commands"""

import json
import unittest
from pathlib import Path

import pytest
import requests_cache
import responses
import ruamel.yaml

import nf_core.modules
import nf_core.modules.create
import nf_core.modules.install
import nf_core.modules.modules_repo
import nf_core.modules.remove
import nf_core.pipelines.create.create
from nf_core import __version__
from nf_core.pipelines.lint_utils import run_prettier_on_file
from nf_core.utils import NFCoreYamlConfig

from .utils import (
    GITLAB_BRANCH_TEST_BRANCH,
    GITLAB_BRANCH_TEST_OLD_SHA,
    GITLAB_DEFAULT_BRANCH,
    GITLAB_URL,
    OLD_TRIMGALORE_BRANCH,
    OLD_TRIMGALORE_SHA,
    create_tmp_pipeline,
    mock_anaconda_api_calls,
    mock_biocontainers_api_calls,
    mock_biotools_api_calls,
)


def create_modules_repo_dummy(tmp_dir):
    """Create a dummy copy of the nf-core/modules repo"""
    yaml = ruamel.yaml.YAML()
    yaml.preserve_quotes = True
    yaml.indent(mapping=2, sequence=2, offset=0)

    root_dir = Path(tmp_dir, "modules")
    Path(root_dir, "modules", "nf-core").mkdir(parents=True)
    Path(root_dir, "tests", "modules", "nf-core").mkdir(parents=True)
    Path(root_dir, "tests", "config").mkdir(parents=True)

    nf_core_yml = NFCoreYamlConfig(nf_core_version=__version__, repository_type="modules", org_path="nf-core")
    with open(Path(root_dir, ".nf-core.yml"), "w") as fh:
        yaml.dump(nf_core_yml.model_dump(), fh)
    # mock biocontainers and anaconda response and biotools response
    with responses.RequestsMock() as rsps:
        mock_anaconda_api_calls(rsps, "bpipe", "0.9.13--hdfd78af_0")
        mock_biocontainers_api_calls(rsps, "bpipe", "0.9.13--hdfd78af_0")
        mock_biotools_api_calls(rsps, "bpipe")
        # bpipe is a valid package on bioconda that is very unlikely to ever be added to nf-core/modules
        module_create = nf_core.modules.create.ModuleCreate(
            root_dir, "bpipe/test", "@author", "process_single", True, False
        )
        with requests_cache.disabled():
            assert module_create.create()

    # Remove doi from meta.yml which makes lint fail
    meta_yml_path = Path(root_dir, "modules", "nf-core", "bpipe", "test", "meta.yml")

    with open(str(meta_yml_path)) as fh:
        meta_yml = yaml.load(fh)
    del meta_yml["tools"][0]["bpipe"]["doi"]
    with open(str(meta_yml_path), "w") as fh:
        yaml.dump(meta_yml, fh)
        run_prettier_on_file(fh.name)
    # Add dummy content to main.nf.test.snap
    test_snap_path = Path(root_dir, "modules", "nf-core", "bpipe", "test", "tests", "main.nf.test.snap")

    with open(test_snap_path, "w") as fh:
        json.dump(
            {
                "my test": {
                    "content": [
                        {
                            "0": [],
                            "versions": {},
                        }
                    ]
                }
            },
            fh,
            indent=4,
        )

    # remove "TODO" statements from main.nf
    main_nf_path = Path(root_dir, "modules", "nf-core", "bpipe", "test", "main.nf")
    with open(main_nf_path) as fh:
        main_nf = fh.read()
    main_nf = main_nf.replace("TODO", "")
    with open(main_nf_path, "w") as fh:
        fh.write(main_nf)

    # remove "TODO" statements from main.nf.test
    main_nf_test_path = Path(root_dir, "modules", "nf-core", "bpipe", "test", "tests", "main.nf.test")
    with open(main_nf_test_path) as fh:
        main_nf_test = fh.read()
    main_nf_test = main_nf_test.replace("TODO", "")
    with open(main_nf_test_path, "w") as fh:
        fh.write(main_nf_test)

    return root_dir


class TestModules(unittest.TestCase):
    """Class for modules tests"""

    def setUp(self):
        """Create a new PipelineSchema and Launch objects"""
        self.component_type = "modules"

        # Set up the schema
        self.tmp_dir, self.template_dir, self.pipeline_name, self.pipeline_dir = create_tmp_pipeline()

        # Set up install objects
        self.mods_install = nf_core.modules.install.ModuleInstall(self.pipeline_dir, prompt=False, force=True)
        self.mods_install_old = nf_core.modules.install.ModuleInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            sha=OLD_TRIMGALORE_SHA,
            remote_url=GITLAB_URL,
            branch=OLD_TRIMGALORE_BRANCH,
        )
        self.mods_install_trimgalore = nf_core.modules.install.ModuleInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            remote_url=GITLAB_URL,
            branch=OLD_TRIMGALORE_BRANCH,
        )
        self.mods_install_gitlab = nf_core.modules.install.ModuleInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            remote_url=GITLAB_URL,
            branch=GITLAB_DEFAULT_BRANCH,
        )
        self.mods_install_gitlab_old = nf_core.modules.install.ModuleInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            remote_url=GITLAB_URL,
            branch=GITLAB_BRANCH_TEST_BRANCH,
            sha=GITLAB_BRANCH_TEST_OLD_SHA,
        )

        # Set up remove objects
        self.mods_remove = nf_core.modules.remove.ModuleRemove(self.pipeline_dir)
        self.mods_remove_gitlab = nf_core.modules.remove.ModuleRemove(
            self.pipeline_dir,
            remote_url=GITLAB_URL,
            branch=GITLAB_DEFAULT_BRANCH,
        )

        # Set up the nf-core/modules repo dummy
        self.nfcore_modules = create_modules_repo_dummy(self.tmp_dir)

    def test_modulesrepo_class(self):
        """Initialise a modules repo object"""
        modrepo = nf_core.modules.modules_repo.ModulesRepo()
        assert modrepo.repo_path == "nf-core"
        assert modrepo.branch == "master"

    @pytest.fixture(autouse=True)
    def _use_caplog(self, caplog):
        self.caplog = caplog



"""Bumps the version number in all appropriate files for
a nf-core pipeline.
"""

import logging
import re
from pathlib import Path
from typing import List, Optional, Tuple, Union

import rich.console
from ruamel.yaml import YAML

import nf_core.utils
from nf_core.pipelines.rocrate import ROCrate
from nf_core.utils import Pipeline

log = logging.getLogger(__name__)
stderr = rich.console.Console(stderr=True, force_terminal=nf_core.utils.rich_force_colors())


def bump_pipeline_version(pipeline_obj: Pipeline, new_version: str) -> None:
    """Bumps a pipeline version number.

    Args:
        pipeline_obj (nf_core.utils.Pipeline): A `Pipeline` object that holds information
            about the pipeline contents and build files.
        new_version (str): The new version tag for the pipeline. Semantic versioning only.
    """

    # Collect the old and new version numbers
    current_version = pipeline_obj.nf_config.get("manifest.version", "").strip(" '\"")
    if new_version.startswith("v"):
        log.warning("Stripping leading 'v' from new version number")
        new_version = new_version[1:]
    if not current_version:
        raise UserWarning("Could not find config variable 'manifest.version'")
    if current_version == new_version:
        raise UserWarning(f"Current version is already: {current_version}")
    log.info(f"Changing version number from '{current_version}' to '{new_version}'")

    # nextflow.config - workflow manifest version
    update_file_version(
        "nextflow.config",
        pipeline_obj,
        [
            (
                rf"(version\s*=\s*['\"]){re.escape(current_version)}(['\"])",
                rf"\g<1>{new_version}\g<2>",
            )
        ],
    )
    # multiqc_config.yaml
    multiqc_new_version = "dev" if "dev" in new_version else new_version
    multiqc_current_version = "dev" if "dev" in current_version else current_version
    if multiqc_current_version != "dev" and multiqc_new_version != "dev":
        update_file_version(
            Path("assets", "multiqc_config.yml"),
            pipeline_obj,
            [
                (
                    f"/releases/tag/{current_version}",
                    f"/releases/tag/{new_version}",
                )
            ],
            yaml_key=["report_comment"],
        )
    if multiqc_current_version != "dev" and multiqc_new_version == "dev":
        update_file_version(
            Path("assets", "multiqc_config.yml"),
            pipeline_obj,
            [
                (
                    f"/releases/tag/{current_version}",
                    "/tree/dev",
                )
            ],
            yaml_key=["report_comment"],
        )
    if multiqc_current_version == "dev" and multiqc_new_version != "dev":
        update_file_version(
            Path("assets", "multiqc_config.yml"),
            pipeline_obj,
            [
                (
                    "/tree/dev",
                    f"/releases/tag/{multiqc_new_version}",
                )
            ],
            yaml_key=["report_comment"],
        )
    update_file_version(
        Path("assets", "multiqc_config.yml"),
        pipeline_obj,
        [
            (
                f"/{multiqc_current_version}/",
                f"/{multiqc_new_version}/",
            ),
        ],
        yaml_key=["report_comment"],
    )
    # nf-test snap files
    pipeline_name = pipeline_obj.nf_config.get("manifest.name", "").strip(" '\"")
    snap_files = [f for f in Path().glob("tests/pipeline/*.snap")]
    for snap_file in snap_files:
        update_file_version(
            snap_file,
            pipeline_obj,
            [
                (
                    f"{pipeline_name}={current_version}",
                    f"{pipeline_name}={new_version}",
                )
            ],
        )
    # .nf-core.yml - pipeline version
    # update entry: version: 1.0.0dev, but not `nf_core_version`, or `bump_version`
    update_file_version(
        ".nf-core.yml",
        pipeline_obj,
        [
            (
                current_version,
                new_version,
            )
        ],
        required=False,
        yaml_key=["template", "version"],
    )

    # update rocrate if ro-crate is present
    if Path(pipeline_obj.wf_path, "ro-crate-metadata.json").exists():
        ROCrate(pipeline_obj.wf_path).update_rocrate()


def bump_nextflow_version(pipeline_obj: Pipeline, new_version: str) -> None:
    """Bumps the required Nextflow version number of a pipeline.

    Args:
        pipeline_obj (nf_core.utils.Pipeline): A `Pipeline` object that holds information
            about the pipeline contents and build files.
        new_version (str): The new version tag for the required Nextflow version.
    """

    # Collect the old and new version numbers - strip leading non-numeric characters (>=)
    current_version = pipeline_obj.nf_config.get("manifest.nextflowVersion", "").strip(" '\"")
    current_version = re.sub(r"^[^0-9\.]*", "", current_version)
    new_version = re.sub(r"^[^0-9\.]*", "", new_version)
    if not current_version:
        raise UserWarning("Could not find config variable 'manifest.nextflowVersion'")
    log.info(f"Changing Nextlow version number from '{current_version}' to '{new_version}'")

    # nextflow.config - manifest minimum nextflowVersion
    update_file_version(
        "nextflow.config",
        pipeline_obj,
        [
            (
                rf"(nextflowVersion\s*=\s*[\'\"]?!>=\s*)({re.escape(current_version)})([\'\"]?)",
                rf"\g<1>{new_version}\g<3>",
            )
        ],
    )

    # .github/workflows/ci.yml - Nextflow version matrix
    update_file_version(
        Path(".github", "workflows", "ci.yml"),
        pipeline_obj,
        [
            (
                # example:
                # NXF_VER:
                #   - "20.04.0"
                current_version,
                new_version,
            )
        ],
        yaml_key=["jobs", "test", "strategy", "matrix", "NXF_VER"],
    )

    # README.md - Nextflow version badge
    update_file_version(
        "README.md",
        pipeline_obj,
        [
            (
                rf"nextflow%20DSL2-%E2%89%A5{re.escape(current_version)}-23aa62.svg",
                f"nextflow%20DSL2-%E2%89%A5{new_version}-23aa62.svg",
            )
        ],
    )


def update_file_version(
    filename: Union[str, Path],
    pipeline_obj: Pipeline,
    patterns: List[Tuple[str, str]],
    required: bool = True,
    yaml_key: Optional[List[str]] = None,
) -> None:
    """
    Updates a file with a new version number.

    Args:
        filename (str): The name of the file to update.
        pipeline_obj (nf_core.utils.Pipeline): A `Pipeline` object that holds information
            about the pipeline contents.
        patterns (List[Tuple[str, str]]): A list of tuples containing the regex patterns to
            match and the replacement strings.
        required (bool, optional): Whether the file is required to exist. Defaults to `True`.
        yaml_key (Optional[List[str]], optional): The YAML key to update. Defaults to `None`.
    """
    fn: Path = pipeline_obj._fp(filename)

    if not fn.exists():
        log.warning(f"File not found: '{fn}'")
        return

    if yaml_key:
        update_yaml_file(fn, patterns, yaml_key, required)
    else:
        update_text_file(fn, patterns, required)


def update_yaml_file(fn: Path, patterns: List[Tuple[str, str]], yaml_key: List[str], required: bool):
    """
    Updates a YAML file with a new version number.

    Args:
        fn (Path): The name of the file to update.
        patterns (List[Tuple[str, str]]): A list of tuples containing the regex patterns to
            match and the replacement strings.
        yaml_key (List[str]): The YAML key to update.
        required (bool): Whether the file is required to exist.
    """
    yaml = YAML()
    yaml.preserve_quotes = True
    with open(fn) as file:
        yaml_content = yaml.load(file)

    try:
        target = yaml_content
        for key in yaml_key[:-1]:
            target = target[key]

        last_key = yaml_key[-1]
        current_value = target[last_key]

        new_value = current_value
        for pattern, replacement in patterns:
            # check if current value is list
            if isinstance(current_value, list):
                new_value = [re.sub(pattern, replacement, item) for item in current_value]
            else:
                new_value = re.sub(pattern, replacement, current_value)

        if new_value != current_value:
            target[last_key] = new_value
            with open(fn, "w") as file:
                yaml.dump(yaml_content, file)
            log.info(f"Updated version in YAML file '{fn}'")
            log_change(str(current_value), str(new_value))
    except KeyError as e:
        handle_error(f"Could not find key {e} in the YAML structure of {fn}", required)


def update_text_file(fn: Path, patterns: List[Tuple[str, str]], required: bool):
    """
    Updates a text file with a new version number.

    Args:
        fn (Path): The name of the file to update.
        patterns (List[Tuple[str, str]]): A list of tuples containing the regex patterns to
            match and the replacement strings.
        required (bool): Whether the file is required to exist.
    """
    with open(fn) as file:
        content = file.read()

    updated = False
    for pattern, replacement in patterns:
        new_content, count = re.subn(pattern, replacement, content)
        if count > 0:
            log_change(content, new_content)
            content = new_content
            updated = True
            log.info(f"Updated version in '{fn}'")
            log.debug(f"Replaced pattern '{pattern}' with '{replacement}' {count} times")
        elif required:
            handle_error(f"Could not find version number in {fn}: `{pattern}`", required)

    if updated:
        with open(fn, "w") as file:
            file.write(content)


def handle_error(message: str, required: bool):
    if required:
        raise ValueError(message)
    else:
        log.info(message)


def log_change(old_content: str, new_content: str):
    old_lines = old_content.splitlines()
    new_lines = new_content.splitlines()

    for old_line, new_line in zip(old_lines, new_lines):
        if old_line != new_line:
            stderr.print(f"          [red] - {old_line.strip()}", highlight=False)
            stderr.print(f"          [green] + {new_line.strip()}", highlight=False)

    stderr.print("\n")



"""A Textual app to create a pipeline."""

from pathlib import Path
from textwrap import dedent

from textual import on
from textual.app import ComposeResult
from textual.containers import Center, Horizontal
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Input, Markdown

from nf_core.pipelines.create.utils import CreateConfig, TextInput, add_hide_class, remove_hide_class

pipeline_exists_warn = """
> ⚠️  **The pipeline you are trying to create already exists.**
>
> If you continue, you will **override** the existing pipeline.
> Please change the pipeline or organisation name to create a different pipeline.
"""


class BasicDetails(Screen):
    """Name, description, author, etc."""

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Markdown(
            dedent(
                """
                # Basic details
                """
            )
        )
        with Horizontal():
            yield TextInput(
                "org",
                "Organisation",
                "GitHub organisation",
                "nf-core",
                classes="column",
                disabled=self.parent.NFCORE_PIPELINE,
            )
            yield TextInput(
                "name",
                "Pipeline Name",
                "Workflow name",
                classes="column",
            )

        yield TextInput(
            "description",
            "Description",
            "A short description of your pipeline.",
        )
        yield TextInput(
            "author",
            "Author(s)",
            "Name of the main author / authors",
        )
        yield Markdown(dedent(pipeline_exists_warn), id="exist_warn", classes="hide")
        yield Center(
            Button("Back", id="back", variant="default"),
            Button("Next", id="next", variant="success"),
            classes="cta",
        )

    @on(Input.Changed)
    @on(Input.Submitted)
    def show_exists_warn(self):
        """Check if the pipeline exists on every input change or submitted.
        If the pipeline exists, show warning message saying that it will be overridden."""
        config = {}
        for text_input in self.query("TextInput"):
            this_input = text_input.query_one(Input)
            config[text_input.field_id] = this_input.value
        if Path(config["org"] + "-" + config["name"]).is_dir():
            remove_hide_class(self.parent, "exist_warn")
        else:
            add_hide_class(self.parent, "exist_warn")

    def on_screen_resume(self):
        """Hide warn message on screen resume.
        Update displayed value on screen resume."""
        add_hide_class(self.parent, "exist_warn")
        for text_input in self.query("TextInput"):
            if text_input.field_id == "org":
                text_input.disabled = self.parent.NFCORE_PIPELINE

    @on(Button.Pressed)
    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Save fields to the config."""
        config = {}
        for text_input in self.query("TextInput"):
            this_input = text_input.query_one(Input)
            validation_result = this_input.validate(this_input.value)
            config[text_input.field_id] = this_input.value
            if not validation_result.is_valid:
                text_input.query_one(".validation_msg").update("\n".join(validation_result.failure_descriptions))
            else:
                text_input.query_one(".validation_msg").update("")
        try:
            self.parent.TEMPLATE_CONFIG = CreateConfig(**config)
            if event.button.id == "next":
                if self.parent.NFCORE_PIPELINE:
                    self.parent.push_screen("type_nfcore")
                else:
                    self.parent.push_screen("type_custom")
        except ValueError:
            pass



import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import questionary
import requests
import rich.prompt

if TYPE_CHECKING:
    from nf_core.modules.modules_repo import ModulesRepo

import nf_core.utils

log = logging.getLogger(__name__)

# Constants for the nf-core/modules repo used throughout the module files
NF_CORE_MODULES_NAME = "nf-core"
NF_CORE_MODULES_REMOTE = "https://github.com/nf-core/modules.git"
NF_CORE_MODULES_DEFAULT_BRANCH = "master"


def get_repo_info(directory: Path, use_prompt: Optional[bool] = True) -> Tuple[Path, Optional[str], str]:
    """
    Determine whether this is a pipeline repository or a clone of
    nf-core/modules
    """

    # Verify that the pipeline dir exists
    if not Path(directory).is_dir():
        raise UserWarning(f"Could not find directory: {directory}")

    # Try to find the root directory
    base_dir: Path = nf_core.utils.determine_base_dir(directory)

    # Figure out the repository type from the .nf-core.yml config file if we can
    config_fn, tools_config = nf_core.utils.load_tools_config(base_dir)
    if config_fn is None:
        raise UserWarning(f"Could not find a config file in directory: {base_dir}")
    repo_type = getattr(tools_config, "repository_type", None) or None

    # If not set, prompt the user
    if not repo_type and use_prompt:
        log.warning("'repository_type' not defined in %s", config_fn.name)
        repo_type = questionary.select(
            "Is this repository a pipeline or a modules repository?",
            choices=[
                {"name": "Pipeline", "value": "pipeline"},
                {"name": "Modules repository", "value": "modules"},
            ],
            style=nf_core.utils.nfcore_question_style,
        ).unsafe_ask()

        # Save the choice in the config file
        log.info(f"To avoid this prompt in the future, add the 'repository_type' key to your {config_fn.name} file.")
        if rich.prompt.Confirm.ask("[bold][blue]?[/] Would you like me to add this config now?", default=True):
            with open(config_fn, "a+") as fh:
                fh.write(f"repository_type: {repo_type}\n")
                log.info(f"Config added to '{config_fn.name}'")

    # Not set and not allowed to ask
    elif not repo_type:
        raise UserWarning("Repository type could not be established")

    # Check if it's a valid answer
    if repo_type not in ["pipeline", "modules"]:
        raise UserWarning(f"Invalid repository type: '{repo_type}', must be 'pipeline' or 'modules'")
    org: str = ""
    # Check for org if modules repo
    if repo_type == "modules":
        org = getattr(tools_config, "org_path", "") or ""
        if org == "":
            log.warning("Organisation path not defined in %s [key: org_path]", config_fn.name)
            org = questionary.text(
                "What is the organisation path under which modules and subworkflows are stored?",
                default="nf-core",
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()
            log.info("To avoid this prompt in the future, add the 'org_path' key to a root '%s' file.", config_fn.name)
            if rich.prompt.Confirm.ask("[bold][blue]?[/] Would you like me to add this config now?", default=True):
                with open(config_fn, "a+") as fh:
                    fh.write(f"org_path: {org}\n")
                    log.info(f"Config added to '{config_fn.name}'")

        if not org:
            raise UserWarning("Organisation path could not be established")

    # It was set on the command line, return what we were given
    return (base_dir, repo_type, org)


def prompt_component_version_sha(
    component_name: str,
    component_type: str,
    modules_repo: "ModulesRepo",
    installed_sha: Optional[str] = None,
) -> str:
    """
    Creates an interactive questionary prompt for selecting the module/subworkflow version
    Args:
        component_name (str): Module/subworkflow name,
        component_type (str): "modules" or "subworkflows",
        modules_repo (ModulesRepo): Modules repo the module/subworkflow originate in
        installed_sha (str): Optional extra argument to highlight the current installed version

    Returns:
        git_sha (str): The selected version of the module/subworkflow
    """
    older_commits_choice = questionary.Choice(
        title=[("fg:ansiyellow", "older commits"), ("class:choice-default", "")], value=""
    )
    git_sha = ""
    page_nbr = 1

    all_commits = iter(modules_repo.get_component_git_log(component_name, component_type))
    next_page_commits = [next(all_commits, None) for _ in range(10)]
    next_page_commits = [commit for commit in next_page_commits if commit is not None]

    while git_sha == "":
        commits = next_page_commits
        next_page_commits = [next(all_commits, None) for _ in range(10)]
        next_page_commits = [commit for commit in next_page_commits if commit is not None]
        if all(commit is None for commit in next_page_commits):
            next_page_commits = []

        choices = []
        for commit in commits:
            if commit:
                title = commit["trunc_message"]
                sha = commit["git_sha"]
                display_color = "fg:ansiblue" if sha != installed_sha else "fg:ansired"
                message = f"{title} {sha}"
                if installed_sha == sha:
                    message += " (installed version)"
                commit_display = [(display_color, message), ("class:choice-default", "")]
                choices.append(questionary.Choice(title=commit_display, value=sha))
        if next_page_commits:
            choices += [older_commits_choice]
        git_sha = questionary.select(
            f"Select '{component_name}' commit:", choices=choices, style=nf_core.utils.nfcore_question_style
        ).unsafe_ask()
        page_nbr += 1
    return git_sha


def get_components_to_install(subworkflow_dir: Union[str, Path]) -> Tuple[List[str], List[str]]:
    """
    Parse the subworkflow main.nf file to retrieve all imported modules and subworkflows.
    """
    modules = []
    subworkflows = []
    with open(Path(subworkflow_dir, "main.nf")) as fh:
        for line in fh:
            regex = re.compile(
                r"include(?: *{ *)([a-zA-Z\_0-9]*)(?: *as *)?(?:[a-zA-Z\_0-9]*)?(?: *})(?: *from *)(?:'|\")(.*)(?:'|\")"
            )
            match = regex.search(line)
            if match and len(match.groups()) == 2:
                name, link = match.groups()
                if link.startswith("../../../"):
                    name_split = name.lower().split("_")
                    modules.append("/".join(name_split))
                elif link.startswith("../"):
                    subworkflows.append(name.lower())
    return modules, subworkflows


def get_biotools_id(tool_name) -> str:
    """
    Try to find a bio.tools ID for 'tool'
    """
    url = f"https://bio.tools/api/t/?q={tool_name}&format=json"
    try:
        # Send a GET request to the API
        response = requests.get(url)
        response.raise_for_status()  # Raise an error for bad status codes
        # Parse the JSON response
        data = response.json()

        # Iterate through the tools in the response to find the tool name
        for tool in data["list"]:
            if tool["name"].lower() == tool_name:
                return tool["biotoolsCURIE"]

        # If the tool name was not found in the response
        log.warning(f"Could not find a bio.tools ID for '{tool_name}'")
        return ""

    except requests.exceptions.RequestException as e:
        log.warning(f"Could not find a bio.tools ID for '{tool_name}': {e}")
        return ""



import os

import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintMergeMarkers(TestLint):
    def test_merge_markers_found(self):
        """Missing 'jobs' field should result in failure"""
        new_pipeline = self._make_pipeline_copy()

        with open(os.path.join(new_pipeline, "main.nf")) as fh:
            main_nf_content = fh.read()
        main_nf_content = ">>>>>>>\n" + main_nf_content
        with open(os.path.join(new_pipeline, "main.nf"), "w") as fh:
            fh.write(main_nf_content)

        lint_obj = nf_core.pipelines.lint.PipelineLint(new_pipeline)
        lint_obj._load()

        results = lint_obj.merge_markers()
        assert len(results["failed"]) > 0
        assert len(results["passed"]) == 0
        assert "Merge marker '>>>>>>>' in " in results["failed"][0]



#!/usr/bin/env python
"""Code to deal with pipeline RO (Research Object) Crates"""

import logging
import os
import re
import sys
from datetime import datetime
from pathlib import Path
from typing import Optional, Set, Union

import requests
import rocrate.rocrate
from git import GitCommandError, InvalidGitRepositoryError
from repo2rocrate.nextflow import NextflowCrateBuilder
from rich.progress import BarColumn, Progress
from rocrate.model.person import Person
from rocrate.rocrate import ROCrate as BaseROCrate

from nf_core.utils import Pipeline

log = logging.getLogger(__name__)


class CustomNextflowCrateBuilder(NextflowCrateBuilder):
    DATA_ENTITIES = NextflowCrateBuilder.DATA_ENTITIES + [
        ("docs/usage.md", "File", "Usage documentation"),
        ("docs/output.md", "File", "Output documentation"),
        ("suborkflows/local", "Dataset", "Pipeline-specific suborkflows"),
        ("suborkflows/nf-core", "Dataset", "nf-core suborkflows"),
        (".nf-core.yml", "File", "nf-core configuration file, configuring template features and linting rules"),
        (".pre-commit-config.yaml", "File", "Configuration file for pre-commit hooks"),
        (".prettierignore", "File", "Ignore file for prettier"),
        (".prettierrc", "File", "Configuration file for prettier"),
    ]


def custom_make_crate(
    root: Path,
    workflow: Optional[Path] = None,
    repo_url: Optional[str] = None,
    wf_name: Optional[str] = None,
    wf_version: Optional[str] = None,
    lang_version: Optional[str] = None,
    ci_workflow: Optional[str] = "ci.yml",
    diagram: Optional[Path] = None,
) -> BaseROCrate:
    builder = CustomNextflowCrateBuilder(root, repo_url=repo_url)

    return builder.build(
        workflow,
        wf_name=wf_name,
        wf_version=wf_version,
        lang_version=lang_version,
        license=None,
        ci_workflow=ci_workflow,
        diagram=diagram,
    )


class ROCrate:
    """
    Class to generate an RO Crate for a pipeline

    """

    def __init__(self, pipeline_dir: Path, version="") -> None:
        """
        Initialise the ROCrate object

        Args:
            pipeline_dir (Path): Path to the pipeline directory
            version (str): Version of the pipeline to checkout
        """
        from nf_core.utils import is_pipeline_directory, setup_requests_cachedir

        is_pipeline_directory(pipeline_dir)
        self.pipeline_dir = pipeline_dir
        self.version: str = version
        self.crate: rocrate.rocrate.ROCrate
        self.pipeline_obj = Pipeline(self.pipeline_dir)
        self.pipeline_obj._load()

        setup_requests_cachedir()

    def create_rocrate(self, json_path: Union[None, Path] = None, zip_path: Union[None, Path] = None) -> bool:
        """
        Create an RO Crate for a pipeline

        Args:
            outdir (Path): Path to the output directory
            json_path (Path): Path to the metadata file
            zip_path (Path): Path to the zip file

        """

        # Check that the checkout pipeline version is the same as the requested version
        if self.version != "":
            if self.version != self.pipeline_obj.nf_config.get("manifest.version"):
                # using git checkout to get the requested version
                log.info(f"Checking out pipeline version {self.version}")
                if self.pipeline_obj.repo is None:
                    log.error(f"Pipeline repository not found in {self.pipeline_dir}")
                    sys.exit(1)
                try:
                    self.pipeline_obj.repo.git.checkout(self.version)
                    self.pipeline_obj = Pipeline(self.pipeline_dir)
                    self.pipeline_obj._load()
                except InvalidGitRepositoryError:
                    log.error(f"Could not find a git repository in {self.pipeline_dir}")
                    sys.exit(1)
                except GitCommandError:
                    log.error(f"Could not checkout version {self.version}")
                    sys.exit(1)
        self.version = self.pipeline_obj.nf_config.get("manifest.version", "")
        self.make_workflow_rocrate()

        # Save just the JSON metadata file
        if json_path is not None:
            if json_path.name == "ro-crate-metadata.json":
                json_path = json_path.parent

            log.info(f"Saving metadata file to '{json_path}'")
            self.crate.metadata.write(json_path)

        # Save the whole crate zip file
        if zip_path is not None:
            if zip_path.name != "ro-crate.crate.zip":
                zip_path = zip_path / "ro-crate.crate.zip"
            log.info(f"Saving zip file '{zip_path}")
            self.crate.write_zip(zip_path)

        if json_path is None and zip_path is None:
            log.error("Please provide a path to save the ro-crate file or the zip file.")
            return False

        return True

    def make_workflow_rocrate(self) -> None:
        """
        Create an RO Crate for a pipeline
        """
        if self.pipeline_obj is None:
            raise ValueError("Pipeline object not loaded")

        diagram: Optional[Path] = None
        # find files (metro|tube)_?(map)?.png in the pipeline directory or docs/ using pathlib
        pattern = re.compile(r".*?(metro|tube|subway)_(map).*?\.png", re.IGNORECASE)
        for file in self.pipeline_dir.rglob("*.png"):
            if pattern.match(file.name):
                log.debug(f"Found diagram: {file}")
                diagram = file.relative_to(self.pipeline_dir)
                break

        # Create the RO Crate object

        self.crate = custom_make_crate(
            self.pipeline_dir,
            self.pipeline_dir / "main.nf",
            self.pipeline_obj.nf_config.get("manifest.homePage", ""),
            self.pipeline_obj.nf_config.get("manifest.name", ""),
            self.pipeline_obj.nf_config.get("manifest.version", ""),
            self.pipeline_obj.nf_config.get("manifest.nextflowVersion", ""),
            diagram=diagram,
        )

        # add readme as description
        readme = self.pipeline_dir / "README.md"

        try:
            self.crate.description = readme.read_text()
        except FileNotFoundError:
            log.error(f"Could not find README.md in {self.pipeline_dir}")
        # get license from LICENSE file
        license_file = self.pipeline_dir / "LICENSE"
        try:
            license = license_file.read_text()
            if license.startswith("MIT"):
                self.crate.license = "MIT"
            else:
                # prompt for license
                log.info("Could not determine license from LICENSE file")
                self.crate.license = input("Please enter the license for this pipeline: ")
        except FileNotFoundError:
            log.error(f"Could not find LICENSE file in {self.pipeline_dir}")

        self.crate.add_jsonld(
            {"@id": "https://nf-co.re/", "@type": "Organization", "name": "nf-core", "url": "https://nf-co.re/"}
        )

        # Set metadata for main entity file
        self.set_main_entity("main.nf")

    def set_main_entity(self, main_entity_filename: str):
        """
        Set the main.nf as the main entity of the crate and add necessary metadata
        """
        if self.crate.mainEntity is None:
            raise ValueError("Main entity not set")

        self.crate.mainEntity.append_to(
            "dct:conformsTo", "https://bioschemas.org/profiles/ComputationalWorkflow/1.0-RELEASE/", compact=True
        )
        # add dateCreated and dateModified, based on the current data
        self.crate.mainEntity.append_to("dateCreated", self.crate.root_dataset.get("dateCreated", ""), compact=True)
        self.crate.mainEntity.append_to(
            "dateModified", str(datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")), compact=True
        )
        self.crate.mainEntity.append_to("sdPublisher", {"@id": "https://nf-co.re/"}, compact=True)
        if self.version.endswith("dev"):
            url = "dev"
        else:
            url = self.version
        self.crate.mainEntity.append_to(
            "url", f"https://nf-co.re/{self.crate.name.replace('nf-core/','')}/{url}/", compact=True
        )
        self.crate.mainEntity.append_to("version", self.version, compact=True)

        # remove duplicate entries for version
        self.crate.mainEntity["version"] = list(set(self.crate.mainEntity["version"]))

        # get keywords from nf-core website
        remote_workflows = requests.get("https://nf-co.re/pipelines.json").json()["remote_workflows"]
        # go through all remote workflows and find the one that matches the pipeline name
        topics = ["nf-core", "nextflow"]
        for remote_wf in remote_workflows:
            assert self.pipeline_obj.pipeline_name is not None  # mypy
            if remote_wf["name"] == self.pipeline_obj.pipeline_name.replace("nf-core/", ""):
                topics = topics + remote_wf["topics"]
                break

        log.debug(f"Adding topics: {topics}")
        self.crate.mainEntity.append_to("keywords", topics)

        self.add_main_authors(self.crate.mainEntity)

        self.crate.mainEntity = self.crate.mainEntity

        self.crate.mainEntity.append_to("license", self.crate.license)
        self.crate.mainEntity.append_to("name", self.crate.name)

        # remove duplicate entries for name
        self.crate.mainEntity["name"] = list(set(self.crate.mainEntity["name"]))

        if "dev" in self.version:
            self.crate.creativeWorkStatus = "InProgress"
        else:
            self.crate.creativeWorkStatus = "Stable"
            if self.pipeline_obj.repo is None:
                log.error(f"Pipeline repository not found in {self.pipeline_dir}")
            else:
                tags = self.pipeline_obj.repo.tags
                if tags:
                    # get the tag for this version
                    for tag in tags:
                        if tag.commit.hexsha == self.pipeline_obj.repo.head.commit.hexsha:
                            self.crate.mainEntity.append_to(
                                "dateCreated",
                                tag.commit.committed_datetime.strftime("%Y-%m-%dT%H:%M:%SZ"),
                                compact=True,
                            )

    def add_main_authors(self, wf_file: rocrate.model.entity.Entity) -> None:
        """
        Add workflow authors to the crate
        """
        # add author entity to crate

        try:
            authors = []
            if "manifest.author" in self.pipeline_obj.nf_config:
                authors.extend([a.strip() for a in self.pipeline_obj.nf_config["manifest.author"].split(",")])
            if "manifest.contributors" in self.pipeline_obj.nf_config:
                contributors = self.pipeline_obj.nf_config["manifest.contributors"]
                names = re.findall(r"name:'([^']+)'", contributors)
                authors.extend(names)
            if not authors:
                raise KeyError("No authors found")
            # add manifest authors as maintainer to crate

        except KeyError:
            log.error("No author or contributors fields found in manifest of nextflow.config")
            return
        # remove duplicates
        authors = list(set(authors))
        # look at git contributors for author names
        try:
            git_contributors: Set[str] = set()
            if self.pipeline_obj.repo is None:
                log.info("No git repository found. No git contributors will be added as authors.")
                return
            commits_touching_path = list(self.pipeline_obj.repo.iter_commits(paths="main.nf"))

            for commit in commits_touching_path:
                if commit.author.name is not None:
                    git_contributors.add(commit.author.name)
            # exclude bots
            contributors = {c for c in git_contributors if not c.endswith("bot") and c != "Travis CI User"}

            log.debug(f"Found {len(contributors)} git authors")

            progress_bar = Progress(
                "[bold blue]{task.description}",
                BarColumn(bar_width=None),
                "[magenta]{task.completed} of {task.total}[reset] » [bold yellow]{task.fields[test_name]}",
                transient=True,
                disable=os.environ.get("HIDE_PROGRESS", None) is not None,
            )
            with progress_bar:
                bump_progress = progress_bar.add_task(
                    "Searching for author names on GitHub", total=len(contributors), test_name=""
                )

                for git_author in contributors:
                    progress_bar.update(bump_progress, advance=1, test_name=git_author)
                    git_author = (
                        requests.get(f"https://api.github.com/users/{git_author}").json().get("name", git_author)
                    )
                    if git_author is None:
                        log.debug(f"Could not find name for {git_author}")
                        continue

        except AttributeError:
            log.debug("Could not find git contributors")

        # remove usernames (just keep names with spaces)
        named_contributors = {c for c in contributors if " " in c}

        for author in named_contributors:
            log.debug(f"Adding author: {author}")

            if self.pipeline_obj.repo is None:
                log.info("No git repository found. No git contributors will be added as authors.")
                return
            # get email from git log
            email = self.pipeline_obj.repo.git.log(f"--author={author}", "--pretty=format:%ae", "-1")
            orcid = get_orcid(author)
            author_entitity = self.crate.add(
                Person(
                    self.crate, orcid if orcid is not None else "#" + email, properties={"name": author, "email": email}
                )
            )
            wf_file.append_to("creator", author_entitity)
            if author in authors:
                wf_file.append_to("maintainer", author_entitity)

    def update_rocrate(self) -> bool:
        """
        Update the rocrate file
        """
        # check if we need to output a json file and/or a zip file based on the file extensions
        # try to find a json file
        json_path: Optional[Path] = None
        potential_json_path = Path(self.pipeline_dir, "ro-crate-metadata.json")
        if potential_json_path.exists():
            json_path = potential_json_path

        # try to find a zip file
        zip_path: Optional[Path] = None
        potential_zip_path = Path(self.pipeline_dir, "ro-crate.crate.zip")
        if potential_zip_path.exists():
            zip_path = potential_zip_path

        return self.create_rocrate(json_path=json_path, zip_path=zip_path)


def get_orcid(name: str) -> Optional[str]:
    """
    Get the ORCID for a given name

    Args:
        name (str): Name of the author

    Returns:
        str: ORCID URI or None
    """
    base_url = "https://pub.orcid.org/v3.0/search/"
    headers = {
        "Accept": "application/json",
    }
    params = {"q": f'family-name:"{name.split()[-1]}" AND given-names:"{name.split()[0]}"'}
    response = requests.get(base_url, params=params, headers=headers)

    if response.status_code == 200:
        json_response = response.json()
        if json_response.get("num-found") == 1:
            orcid_uri = json_response.get("result")[0].get("orcid-identifier", {}).get("uri")
            log.info(f"Using found ORCID for {name}. Please double-check: {orcid_uri}")
            return orcid_uri
        else:
            log.debug(f"No exact ORCID found for {name}. See {response.url}")
            return None
    else:
        log.info(f"API request to ORCID unsuccessful. Status code: {response.status_code}")
        return None



from ..test_lint import TestLint


class TestLintModulesJson(TestLint):
    def test_modules_json_pass(self):
        self.lint_obj._load()
        results = self.lint_obj.modules_json()
        assert len(results.get("warned", [])) == 0
        assert len(results.get("failed", [])) == 0
        assert len(results.get("passed", [])) > 0



from rich.console import Console

import nf_core.subworkflows

from ..test_subworkflows import TestSubworkflows
from ..utils import GITLAB_SUBWORKFLOWS_BRANCH, GITLAB_URL


class TestSubworkflowsList(TestSubworkflows):
    def test_subworkflows_list_remote(self):
        """Test listing available subworkflows"""
        subworkflows_list = nf_core.subworkflows.SubworkflowList(remote=True)
        listed_subworkflows = subworkflows_list.list_components()
        console = Console(record=True)
        console.print(listed_subworkflows)
        output = console.export_text()
        assert "bam_stats" in output

    def test_subworkflows_list_remote_gitlab(self):
        """Test listing the subworkflows in the remote gitlab repo"""
        subworkflows_list = nf_core.subworkflows.SubworkflowList(
            remote=True, remote_url=GITLAB_URL, branch=GITLAB_SUBWORKFLOWS_BRANCH
        )
        listed_subworkflows = subworkflows_list.list_components()
        console = Console(record=True)
        console.print(listed_subworkflows)
        output = console.export_text()
        assert "bam_stats" in output

    def test_subworkflows_install_and_list_subworkflows(self):
        """Test listing locally installed subworkflows"""
        self.subworkflow_install.install("bam_sort_stats_samtools")
        subworkflows_list = nf_core.subworkflows.SubworkflowList(self.pipeline_dir, remote=False)
        listed_subworkflows = subworkflows_list.list_components()
        console = Console(record=True)
        console.print(listed_subworkflows)
        output = console.export_text()
        assert "bam_stats" in output

    def test_subworkflows_install_gitlab_and_list_subworkflows(self):
        """Test listing locally installed subworkflows"""
        self.subworkflow_install_gitlab.install("bam_sort_stats_samtools")
        subworkflows_list = nf_core.subworkflows.SubworkflowList(self.pipeline_dir, remote=False)
        listed_subworkflows = subworkflows_list.list_components()
        console = Console(record=True)
        console.print(listed_subworkflows)
        output = console.export_text()
        assert "bam_stats" in output



from textual.app import ComposeResult
from textual.containers import Center, Grid
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Markdown

markdown_intro = """
# Choose pipeline type
"""

markdown_type_nfcore = """
## Choose _"nf-core"_ if:

* You want your pipeline to be part of the nf-core community
* You think that there's an outside chance that it ever _could_ be part of nf-core
"""
markdown_type_custom = """
## Choose _"Custom"_ if:

* Your pipeline will _never_ be part of nf-core
* You want full control over *all* features that are included from the template
      (including those that are mandatory for nf-core).
"""

markdown_details = """
## What's the difference?

Choosing _"nf-core"_ effectively pre-selects the following template features:

* GitHub Actions continuous-integration configuration files:
    * Pipeline test runs: Small-scale (GitHub) and large-scale (AWS)
    * Code formatting checks with [Prettier](https://prettier.io/)
    * Auto-fix linting functionality using [@nf-core-bot](https://github.com/nf-core-bot)
    * Marking old issues as stale
* Inclusion of [shared nf-core configuration profiles](https://nf-co.re/configs)
"""


class ChoosePipelineType(Screen):
    """Choose whether this will be an nf-core pipeline or not."""

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Markdown(markdown_intro)
        yield Grid(
            Center(
                Markdown(markdown_type_nfcore),
                Center(Button("nf-core", id="type_nfcore", variant="success")),
            ),
            Center(
                Markdown(markdown_type_custom),
                Center(Button("Custom", id="type_custom", variant="primary")),
            ),
            classes="col-2 pipeline-type-grid",
        )
        yield Markdown(markdown_details)



import logging
import os
import sys
from pathlib import Path
from typing import Optional, Union

import rich

from nf_core.pipelines.params_file import ParamsFileBuilder
from nf_core.utils import rich_force_colors

log = logging.getLogger(__name__)

stdout = rich.console.Console(force_terminal=rich_force_colors())

## nf-core pipelines command functions ##


# nf-core pipelines create
def pipelines_create(ctx, name, description, author, version, force, outdir, template_yaml, organisation):
    """
    Create a new pipeline using the nf-core template.

    Uses the nf-core template to make a skeleton Nextflow pipeline with all required
    files, boilerplate code and best-practices.
    \n\n
    Run without any command line arguments to use an interactive interface.
    """
    from nf_core.pipelines.create import PipelineCreateApp
    from nf_core.pipelines.create.create import PipelineCreate

    if (name and description and author) or (template_yaml):
        # If all command arguments are used, run without the interactive interface
        try:
            create_obj = PipelineCreate(
                name,
                description,
                author,
                version=version,
                force=force,
                outdir=outdir,
                template_config=template_yaml,
                organisation=organisation,
            )
            create_obj.init_pipeline()
        except UserWarning as e:
            log.error(e)
            sys.exit(1)
    elif name or description or author or version != "1.0.0dev" or force or outdir or organisation != "nf-core":
        log.error(
            "[red]Partial arguments supplied.[/] "
            "Run without [i]any[/] arguments for an interactive interface, "
            "or with at least name + description + author to use non-interactively."
        )
        sys.exit(1)
    else:
        log.info("Launching interactive nf-core pipeline creation tool.")
        app = PipelineCreateApp()
        app.run()
        sys.exit(app.return_code or 0)


# nf-core pipelines bump-version
def pipelines_bump_version(ctx, new_version, directory, nextflow):
    """
    Update nf-core pipeline version number.

    The pipeline version number is mentioned in a lot of different places
    in nf-core pipelines. This tool updates the version for you automatically,
    so that you don't accidentally miss any.

    Should be used for each pipeline release, and again for the next
    development version after release.

    As well as the pipeline version, you can also change the required version of Nextflow.
    """
    from nf_core.pipelines.bump_version import bump_nextflow_version, bump_pipeline_version
    from nf_core.utils import Pipeline, is_pipeline_directory

    try:
        # Check if pipeline directory contains necessary files
        is_pipeline_directory(directory)

        # Make a pipeline object and load config etc
        pipeline_obj = Pipeline(directory)
        pipeline_obj._load()

        # Bump the pipeline version number
        if not nextflow:
            bump_pipeline_version(pipeline_obj, new_version)
        else:
            bump_nextflow_version(pipeline_obj, new_version)
    except UserWarning as e:
        log.error(e)
        sys.exit(1)


# nf-core pipelines lint
def pipelines_lint(
    ctx,
    directory,
    release,
    fix,
    key,
    show_passed,
    fail_ignored,
    fail_warned,
    markdown,
    json,
    sort_by,
):
    """
    Check pipeline code against nf-core guidelines.

    Runs a large number of automated tests to ensure that the supplied pipeline
    meets the nf-core guidelines. Documentation of all lint tests can be found
    on the nf-core website: [link=https://nf-co.re/tools/docs/]https://nf-co.re/tools/docs/[/]

    You can ignore tests using a file called [blue].nf-core.yml[/] [i](if you have a good reason!)[/].
    See the documentation for details.
    """
    from nf_core.pipelines.lint import run_linting
    from nf_core.utils import is_pipeline_directory

    # Check if pipeline directory is a pipeline
    try:
        is_pipeline_directory(directory)
    except UserWarning as e:
        log.error(e)
        sys.exit(1)

    # Run the lint tests!
    try:
        lint_obj, module_lint_obj, subworkflow_lint_obj = run_linting(
            directory,
            release,
            fix,
            key,
            show_passed,
            fail_ignored,
            fail_warned,
            sort_by,
            markdown,
            json,
            ctx.obj["hide_progress"],
        )
        swf_failed = 0
        module_failed = 0
        if subworkflow_lint_obj is not None:
            swf_failed = len(subworkflow_lint_obj.failed)
        if module_lint_obj is not None:
            module_failed = len(module_lint_obj.failed)
        if len(lint_obj.failed) + module_failed + swf_failed > 0:
            sys.exit(1)
    except AssertionError as e:
        log.critical(e)
        sys.exit(1)
    except UserWarning as e:
        log.error(e)
        sys.exit(1)


# nf-core pipelines download
def pipelines_download(
    ctx,
    pipeline,
    revision,
    outdir,
    compress,
    force,
    platform,
    download_configuration,
    tag,
    container_system,
    container_library,
    container_cache_utilisation,
    container_cache_index,
    parallel_downloads,
):
    """
    Download a pipeline, nf-core/configs and pipeline singularity images.

    Collects all files in a single archive and configures the downloaded
    workflow to use relative paths to the configs and singularity images.
    """
    from nf_core.pipelines.download import DownloadWorkflow

    dl = DownloadWorkflow(
        pipeline,
        revision,
        outdir,
        compress,
        force,
        platform,
        download_configuration,
        tag,
        container_system,
        container_library,
        container_cache_utilisation,
        container_cache_index,
        parallel_downloads,
    )
    dl.download_workflow()


# nf-core pipelines create-params-file
def pipelines_create_params_file(ctx, pipeline, revision, output, force, show_hidden):
    """
    Build a parameter file for a pipeline.

    Uses the pipeline schema file to generate a YAML parameters file.
    Parameters are set to the pipeline defaults and descriptions are shown in comments.
    After the output file is generated, it can then be edited as needed before
    passing to nextflow using the `-params-file` option.

    Run using a remote pipeline name (such as GitHub `user/repo` or a URL),
    a local pipeline directory.
    """
    builder = ParamsFileBuilder(pipeline, revision)

    if not builder.write_params_file(output, show_hidden=show_hidden, force=force):
        sys.exit(1)


# nf-core pipelines launch
def pipelines_launch(
    ctx,
    pipeline,
    id,
    revision,
    command_only,
    params_in,
    params_out,
    save_all,
    show_hidden,
    url,
):
    """
    Launch a pipeline using a web GUI or command line prompts.

    Uses the pipeline schema file to collect inputs for all available pipeline
    parameters. Parameter names, descriptions and help text are shown.
    The pipeline schema is used to validate all inputs as they are entered.

    When finished, saves a file with the selected parameters which can be
    passed to Nextflow using the -params-file option.

    Run using a remote pipeline name (such as GitHub `user/repo` or a URL),
    a local pipeline directory or an ID from the nf-core web launch tool.
    """
    from nf_core.pipelines.launch import Launch

    launcher = Launch(
        pipeline,
        revision,
        command_only,
        params_in,
        params_out,
        save_all,
        show_hidden,
        url,
        id,
    )
    if not launcher.launch_pipeline():
        sys.exit(1)


# nf-core pipelines list
def pipelines_list(ctx, keywords, sort, json, show_archived):
    """
    List available nf-core pipelines with local info.

    Checks the web for a list of nf-core pipelines with their latest releases.
    Shows which nf-core pipelines you have pulled locally and whether they are up to date.
    """
    from nf_core.pipelines.list import list_workflows

    stdout.print(list_workflows(keywords, sort, json, show_archived))


# nf-core pipelines rocrate
def pipelines_rocrate(
    ctx,
    pipeline_dir: Union[str, Path],
    json_path: Optional[Union[str, Path]],
    zip_path: Optional[Union[str, Path]],
    pipeline_version: str,
) -> None:
    from nf_core.pipelines.rocrate import ROCrate

    if json_path is None and zip_path is None:
        log.error("Either `--json_path` or `--zip_path` must be specified.")
        sys.exit(1)
    else:
        pipeline_dir = Path(pipeline_dir)
        if json_path is not None:
            json_path = Path(json_path)
        if zip_path is not None:
            zip_path = Path(zip_path)
        try:
            rocrate_obj = ROCrate(pipeline_dir, pipeline_version)
            rocrate_obj.create_rocrate(json_path=json_path, zip_path=zip_path)
        except (UserWarning, LookupError, FileNotFoundError) as e:
            log.error(e)
            sys.exit(1)


# nf-core pipelines sync
def pipelines_sync(ctx, directory, from_branch, pull_request, github_repository, username, template_yaml, force_pr):
    """
    Sync a pipeline [cyan i]TEMPLATE[/] branch with the nf-core template.

    To keep nf-core pipelines up to date with improvements in the main
    template, we use a method of synchronisation that uses a special
    git branch called [cyan i]TEMPLATE[/].

    This command updates the [cyan i]TEMPLATE[/] branch with the latest version of
    the nf-core template, so that these updates can be synchronised with
    the pipeline. It is run automatically for all pipelines when ever a
    new release of [link=https://github.com/nf-core/tools]nf-core/tools[/link] (and the included template) is made.
    """
    from nf_core.pipelines.sync import PipelineSync, PullRequestExceptionError, SyncExceptionError
    from nf_core.utils import is_pipeline_directory

    try:
        # Check if pipeline directory contains necessary files
        is_pipeline_directory(directory)
        # Sync the given pipeline dir
        sync_obj = PipelineSync(
            directory, from_branch, pull_request, github_repository, username, template_yaml, force_pr
        )
        sync_obj.sync()
    except (SyncExceptionError, PullRequestExceptionError) as e:
        log.error(e)
        sys.exit(1)


# nf-core pipelines create-logo
def pipelines_create_logo(logo_text, directory, name, theme, width, format, force):
    """
    Generate a logo with the nf-core logo template.

    This command generates an nf-core pipeline logo, using the supplied <logo_text>
    """
    from nf_core.pipelines.create_logo import create_logo

    try:
        if directory == ".":
            directory = Path.cwd()
        logo_path = create_logo(logo_text, directory, name, theme, width, format, force)
        # Print path to logo relative to current working directory
        try:
            logo_path = Path(logo_path).relative_to(Path.cwd())
        except ValueError:
            logo_path = Path(logo_path)
        log.info(f"Created logo: [magenta]{logo_path}[/]")
    except UserWarning as e:
        log.error(e)
        sys.exit(1)


# nf-core pipelines schema validate
def pipelines_schema_validate(pipeline, params):
    """
    Validate a set of parameters against a pipeline schema.

    Nextflow can be run using the -params-file flag, which loads
    script parameters from a JSON file.

    This command takes such a file and validates it against the pipeline
    schema, checking whether all schema rules are satisfied.
    """
    from nf_core.pipelines.schema import PipelineSchema

    schema_obj = PipelineSchema()
    try:
        schema_obj.get_schema_path(pipeline)
        # Load and check schema
        schema_obj.load_lint_schema()
    except AssertionError as e:
        log.error(e)
        sys.exit(1)
    schema_obj.load_input_params(params)
    try:
        schema_obj.validate_params()
    except AssertionError:
        sys.exit(1)


# nf-core pipelines schema build
def pipelines_schema_build(directory, no_prompts, web_only, url):
    """
    Interactively build a pipeline schema from Nextflow params.

    Automatically detects parameters from the pipeline config and main.nf and
    compares these to the pipeline schema. Prompts to add or remove parameters
    if the two do not match one another.

    Once all parameters are accounted for, can launch a web GUI tool on the
    https://nf-co.re website where you can annotate and organise parameters.
    Listens for this to be completed and saves the updated schema.
    """
    from nf_core.pipelines.schema import PipelineSchema

    try:
        schema_obj = PipelineSchema()
        if schema_obj.build_schema(directory, no_prompts, web_only, url) is False:
            sys.exit(1)
    except (UserWarning, AssertionError) as e:
        log.error(e)
        sys.exit(1)


# nf-core pipelines schema lint
def pipelines_schema_lint(schema_path):
    """
    Check that a given pipeline schema is valid.

    Checks whether the pipeline schema validates as JSON Schema Draft 7
    and adheres to the additional nf-core pipelines schema requirements.

    This function runs as part of the nf-core pipelines lint command, this is a convenience
    command that does just the schema linting nice and quickly.

    If no schema path is provided, "nextflow_schema.json" will be used (if it exists).
    """
    from nf_core.pipelines.schema import PipelineSchema

    schema_obj = PipelineSchema()
    try:
        schema_obj.get_schema_path(schema_path)
        schema_obj.load_lint_schema()
        # Validate title and description - just warnings as schema should still work fine
        try:
            schema_obj.validate_schema_title_description()
        except AssertionError as e:
            log.warning(e)
    except AssertionError:
        sys.exit(1)


# nf-core pipelines schema docs
def pipelines_schema_docs(schema_path, output, format, force, columns):
    """
    Outputs parameter documentation for a pipeline schema.
    """
    if not os.path.exists(schema_path):
        log.error("Could not find 'nextflow_schema.json' in current directory. Please specify a path.")
        sys.exit(1)

    from nf_core.pipelines.schema import PipelineSchema

    schema_obj = PipelineSchema()
    # Assume we're in a pipeline dir root if schema path not set
    schema_obj.get_schema_path(schema_path)
    schema_obj.load_schema()
    schema_obj.print_documentation(output, format, force, columns.split(","))



import logging
from pathlib import Path
from typing import Optional, Union

from PIL import Image, ImageDraw, ImageFont

import nf_core
from nf_core.utils import NFCORE_CACHE_DIR

log = logging.getLogger(__name__)


def create_logo(
    text: str,
    directory: Union[Path, str],
    filename: str = "",
    theme: str = "light",
    width: int = 2300,
    format: str = "png",
    force: bool = False,
) -> Path:
    """Create a logo for a pipeline."""
    if not text:
        raise UserWarning("Please provide the name of the text to put on the logo.")
    directory = Path(directory)
    if not directory.is_dir():
        log.debug(f"Creating directory {directory}")
        directory.mkdir(parents=True, exist_ok=True)
    assets = Path(nf_core.__file__).parent / "assets/logo"

    if format == "svg":
        template_fn = "placeholder_logo.svg"

        if width != 2300:
            log.warning("SVG format does not support resizing. Setting width to 2300px.")

        # replace the placeholder text with the pipeline name
        with open(assets / template_fn) as fh:
            svg = fh.read().replace("PLACEHOLDER", text)
        if theme == "dark":
            svg = svg.replace("#050505", "#fafafa")

        # save the svg
        logo_filename = f"nf-core-{text}_logo_{theme}.svg" if not filename else filename
        logo_filename = f"{logo_filename}.svg" if not logo_filename.lower().endswith(".svg") else logo_filename
        logo_path = Path(directory, logo_filename)
        with open(logo_path, "w") as fh:
            fh.write(svg)

    else:
        logo_filename = f"nf-core-{text}_logo_{theme}.png" if not filename else filename
        logo_filename = f"{logo_filename}.png" if not logo_filename.lower().endswith(".png") else logo_filename
        cache_name = f"nf-core-{text}_logo_{theme}_{width}.png"
        logo_path = Path(directory, logo_filename)

        # Check if we haven't already created this logo
        if logo_path.is_file() and not force:
            log.info(f"Logo already exists at: {logo_path}. Use `--force` to overwrite.")
            return logo_path
        # cache file
        cache_path = Path(NFCORE_CACHE_DIR, "logo", cache_name)
        img: Optional[Image.Image] = None
        if cache_path.is_file():
            log.debug(f"Logo already exists in cache at: {cache_path}. Reusing this file.")
            img = Image.open(cache_path)
        if img is None:
            log.debug(f"Creating logo for {text}")

            # make sure the figure fits the text
            font_path = assets / "MavenPro-Bold.ttf"
            log.debug(f"Using font: {str(font_path)}")
            font = ImageFont.truetype(str(font_path), 400)
            text_length = font.getmask(text).getbbox()[2]  # get the width of the text based on the font

            max_width = max(
                2300, text_length + len(text) * 20
            )  # need to add some more space to the text length to make sure it fits

            template_fn = "nf-core-repo-logo-base-lightbg.png"
            if theme == "dark":
                template_fn = "nf-core-repo-logo-base-darkbg.png"

            template_path = assets / template_fn
            img = Image.open(template_path)
            # get the height of the template image
            height = img.size[1]

            # Draw text
            draw = ImageDraw.Draw(img)
            color = theme == "dark" and (250, 250, 250) or (5, 5, 5)
            draw.text((110, 465), text, color, font=font)

            if img is not None:
                # Crop to max width
                img = img.crop((0, 0, max_width, height))

                # Resize
                img = img.resize((width, int((width / max_width) * height)))
            else:
                log.error("Failed to create logo, no image object created.")

            # Save to cache
            Path(cache_path.parent).mkdir(parents=True, exist_ok=True)
            log.debug(f"Saving logo to cache: {cache_path}")
            img.save(cache_path, "PNG")
        # Save
        img.save(logo_path, "PNG")

    log.debug(f"Saved logo to: '{logo_path}'")

    # Return the logo
    return logo_path



import logging

from nf_core.components.info import ComponentInfo

log = logging.getLogger(__name__)


class ModuleInfo(ComponentInfo):
    def __init__(
        self,
        pipeline_dir,
        component_name,
        remote_url=None,
        branch=None,
        no_pull=False,
    ):
        super().__init__("modules", pipeline_dir, component_name, remote_url, branch, no_pull)



"""Tests covering the command-line code.

Most tests check the cli arguments are passed along and that some action is
taken.
"""

import tempfile
import unittest
from pathlib import Path
from unittest import mock

from click.testing import CliRunner

import nf_core.__main__
import nf_core.utils


@mock.patch("nf_core.__main__.nf_core_cli")
def test_header(mock_cli):
    """Just try to execute the header function"""
    nf_core.__main__.run_nf_core()


@mock.patch("nf_core.__main__.nf_core_cli")
@mock.patch("nf_core.__main__.check_if_outdated", return_value=(True, None, "dummy_version"))
def test_header_outdated(mock_check_outdated, mock_nf_core_cli, capsys):
    """Check cli notifies the user when nf_core is outdated"""
    nf_core.__main__.run_nf_core()
    captured = capsys.readouterr()
    assert "There is a new version of nf-core/tools available! (dummy_version)" in captured.err


class TestCli(unittest.TestCase):
    """Class for testing the command line interface"""

    def setUp(self):
        self.runner = CliRunner()

    def assemble_params(self, params):
        """Assemble a dictionary of parameters into a list of arguments for the cli

        Note:
            if the value of a parameter is None, it will be considered a flag.
            Booleans were not used to avoid conflicting with the click.BOOL type.

        Args:
            params (dict): dict of parameters to assemble"""
        arg_list = []
        for key, value in params.items():
            if value is not None:
                arg_list += [f"--{key}", value]
            else:
                arg_list += [f"--{key}"]

        return arg_list

    def invoke_cli(self, cmd):
        """Invoke the commandline interface using a list of parameters

        Args:
            cmd (list): commandline to execute
        """
        return self.runner.invoke(nf_core.__main__.nf_core_cli, cmd)

    def test_cli_help(self):
        """Test the main launch function with --help"""
        result = self.invoke_cli(["--help"])
        assert result.exit_code == 0
        assert "Show the version and exit." in result.output

    def test_cli_bad_subcommand(self):
        """Test the main launch function with an unrecognised argument"""
        result = self.invoke_cli(["foo"])
        assert result.exit_code == 2

    def test_cli_verbose(self):
        """Test the main launch function with verbose flag"""
        result = self.invoke_cli(["-v"])
        # Checks that -v was considered valid
        assert "No such option: -v" not in nf_core.utils.strip_ansi_codes(result.output)

    @mock.patch("nf_core.pipelines.list.list_workflows", return_value="pipeline test list")
    def test_cli_list(self, mock_list_workflows):
        """Test nf-core pipelines are listed and cli parameters are passed on."""
        params = {
            "sort": "name",
            "json": None,
            "show-archived": None,
        }
        cmd = ["pipelines", "list"] + self.assemble_params(params) + ["kw1", "kw2"]
        result = self.invoke_cli(cmd)

        mock_list_workflows.assert_called_once_with(
            tuple(cmd[-2:]), params["sort"], "json" in params, "show-archived" in params
        )
        assert result.exit_code == 0
        assert "pipeline test list" in result.output

    @mock.patch("nf_core.pipelines.launch.Launch")
    def test_cli_launch(self, mock_launcher):
        """Test nf-core pipeline is launched and cli parameters are passed on."""
        mock_launcher.return_value.launch_pipeline.return_value = True

        temp_params_in = tempfile.NamedTemporaryFile()
        params = {
            "revision": "abcdef",
            "id": "idgui",
            "command-only": None,
            "params-out": "/path/params/out",
            "params-in": temp_params_in.name,
            "save-all": None,
            "show-hidden": None,
            "url": "builder_url",
        }
        cmd = ["pipelines", "launch"] + self.assemble_params(params) + ["pipeline_name"]
        result = self.invoke_cli(cmd)

        assert result.exit_code == 0

        mock_launcher.assert_called_once_with(
            cmd[-1],
            params["revision"],
            "command-only" in params,
            params["params-in"],
            params["params-out"],
            "save-all" in params,
            "show-hidden" in params,
            params["url"],
            params["id"],
        )

        mock_launcher.return_value.launch_pipeline.assert_called_once()

    @mock.patch("nf_core.pipelines.launch.Launch")
    def test_cli_launch_no_params_in(self, mock_launcher):
        """Test nf-core pipeline fails when params-in does not exist"""
        mock_launcher.return_value.launch_pipeline.return_value = True

        params = {
            "params-in": "/fake/path",
        }
        cmd = ["pipelines", "launch"] + self.assemble_params(params) + ["pipeline_name"]
        result = self.invoke_cli(cmd)

        assert result.exit_code == 2
        assert (
            f"Invalid value for '-p' / '--params-in': Path '{params['params-in']}' does not exist."
            in nf_core.utils.strip_ansi_codes(result.output)
        )

        mock_launcher.assert_not_called()

    @mock.patch("nf_core.pipelines.launch.Launch")
    def test_cli_launch_fail(self, mock_launcher):
        """Test nf-core pipeline fails with exit code 1 when pipeline fails."""
        mock_launcher.return_value.launch_pipeline.return_value = False
        cmd = ["pipelines", "launch", "pipeline_name"]
        result = self.invoke_cli(cmd)
        assert result.exit_code == 1

    @mock.patch("nf_core.pipelines.download.DownloadWorkflow")
    def test_cli_download(self, mock_dl):
        """Test nf-core pipeline is downloaded and cli parameters are passed on."""
        params = {
            "revision": "abcdef",
            "outdir": "/path/outdir",
            "compress": "tar.gz",
            "force": None,
            "platform": None,
            "download-configuration": "yes",
            "tag": "3.12=testing",
            "container-system": "singularity",
            "container-library": "quay.io",
            "container-cache-utilisation": "copy",
            "container-cache-index": "/path/index.txt",
            "parallel-downloads": 2,
        }

        cmd = ["pipelines", "download"] + self.assemble_params(params) + ["pipeline_name"]
        result = self.invoke_cli(cmd)

        assert result.exit_code == 0

        mock_dl.assert_called_once_with(
            cmd[-1],
            (params["revision"],),
            params["outdir"],
            params["compress"],
            "force" in params,
            "platform" in params,
            params["download-configuration"],
            (params["tag"],),
            params["container-system"],
            (params["container-library"],),
            params["container-cache-utilisation"],
            params["container-cache-index"],
            params["parallel-downloads"],
        )

        mock_dl.return_value.download_workflow.assert_called_once()

    @mock.patch("nf_core.pipelines.create.create.PipelineCreate")
    def test_create(self, mock_create):
        """Test nf-core pipeline is created and cli parameters are passed on."""
        params = {
            "name": "pipelinename",
            "description": "pipeline description",
            "author": "Kalle Anka",
            "outdir": "/path/outdir",
        }

        cmd = ["pipelines", "create"] + self.assemble_params(params)
        result = self.invoke_cli(cmd)

        assert result.exit_code == 0
        mock_create.assert_called_once_with(
            params["name"],
            params["description"],
            params["author"],
            force="force" in params,
            version="1.0.0dev",
            outdir=params["outdir"],
            template_config=None,
            organisation="nf-core",
        )
        mock_create.return_value.init_pipeline.assert_called_once()

    @mock.patch("nf_core.pipelines.create.create.PipelineCreate")
    def test_create_error(self, mock_create):
        """Test `nf-core pipelines create` run without providing all the arguments thorws an error."""
        params = {
            "name": "pipelinename",
        }

        cmd = ["pipelines", "create"] + self.assemble_params(params)
        result = self.invoke_cli(cmd)

        assert result.exit_code == 1
        assert "Partial arguments supplied." in result.output

    @mock.patch("nf_core.pipelines.create.PipelineCreateApp")
    def test_create_app(self, mock_create):
        """Test `nf-core pipelines create` runs an App."""
        cmd = ["pipelines", "create"]
        result = self.invoke_cli(cmd)

        assert result.return_value == (0 or None)
        assert "Launching interactive nf-core pipeline creation tool." in result.output

        mock_create.assert_called_once_with()
        mock_create.return_value.run.assert_called_once()

    @mock.patch("nf_core.utils.is_pipeline_directory")
    @mock.patch("nf_core.pipelines.lint.run_linting")
    def test_lint(self, mock_lint, mock_is_pipeline):
        """Test nf-core pipelines lint"""
        mock_lint_results = (mock.MagicMock, mock.MagicMock, mock.MagicMock)
        mock_lint_results[0].failed = []
        mock_lint_results[1].failed = []
        mock_lint_results[2].failed = []
        mock_lint.return_value = mock_lint_results

        temp_pipeline_dir = tempfile.NamedTemporaryFile()
        params = {
            "dir": temp_pipeline_dir.name,
            "release": None,
            "fix": "fix test",
            "key": "key test",
            "show-passed": None,
            "fail-ignored": None,
            "fail-warned": None,
            "markdown": "output_file.md",
            "json": "output_file.json",
        }

        cmd = ["pipelines", "lint"] + self.assemble_params(params)
        result = self.invoke_cli(cmd)

        assert result.exit_code == 0
        mock_lint.assert_called_once_with(
            params["dir"],
            "release" in params,
            (params["fix"],),
            (params["key"],),
            "show-passed" in params,
            "fail-ignored" in params,
            "fail-warned" in params,
            "test",
            params["markdown"],
            params["json"],
            "hide-progress" in params,
        )

    def test_lint_no_dir(self):
        """Test nf-core pipelines lint fails if --dir does not exist"""
        params = {
            "dir": "/bad/path",
        }

        cmd = ["pipelines", "lint"] + self.assemble_params(params)
        result = self.invoke_cli(cmd)

        assert result.exit_code == 2
        assert (
            f"Invalid value for '-d' / '--dir': Path '{params['dir']}' does not exist."
            in nf_core.utils.strip_ansi_codes(result.output)
        )

    @mock.patch("nf_core.utils.is_pipeline_directory")
    def test_lint_dir_is_not_pipeline(self, mock_is_pipeline):
        """Test nf-core pipelines lint logs an error if not called from a pipeline directory."""
        error_txt = "UserWarning has been raised"
        mock_is_pipeline.side_effect = UserWarning(error_txt)

        cmd = ["pipelines", "lint"]
        with self.assertLogs() as captured_logs:
            result = self.invoke_cli(cmd)

        assert result.exit_code == 1
        assert error_txt in captured_logs.output[-1]
        assert captured_logs.records[-1].levelname == "ERROR"

    @mock.patch("nf_core.utils.is_pipeline_directory")
    @mock.patch("nf_core.pipelines.lint.run_linting")
    def test_lint_log_assert_error(self, mock_lint, mock_is_pipeline):
        """Test nf-core pipelines lint logs assertion errors"""
        error_txt = "AssertionError has been raised"
        mock_lint.side_effect = AssertionError(error_txt)

        cmd = ["pipelines", "lint"]
        with self.assertLogs() as captured_logs:
            result = self.invoke_cli(cmd)

        assert result.exit_code == 1
        assert error_txt in captured_logs.output[-1]
        assert captured_logs.records[-1].levelname == "CRITICAL"

    @mock.patch("nf_core.utils.is_pipeline_directory")
    @mock.patch("nf_core.pipelines.lint.run_linting")
    def test_lint_log_user_warning(self, mock_lint, mock_is_pipeline):
        """Test nf-core pipelines lint logs assertion errors"""
        error_txt = "AssertionError has been raised"
        mock_lint.side_effect = UserWarning(error_txt)

        cmd = ["pipelines", "lint"]
        with self.assertLogs() as captured_logs:
            result = self.invoke_cli(cmd)

        assert result.exit_code == 1
        assert error_txt in captured_logs.output[-1]
        assert captured_logs.records[-1].levelname == "ERROR"

    @mock.patch("nf_core.pipelines.schema.PipelineSchema.get_schema_path")
    def test_schema_lint(self, mock_get_schema_path):
        """Test nf-core pipelines schema lint defaults to nextflow_schema.json"""
        cmd = ["pipelines", "schema", "lint"]
        with self.runner.isolated_filesystem():
            with open("nextflow_schema.json", "w") as f:
                f.write("{}")
            self.invoke_cli(cmd)
            mock_get_schema_path.assert_called_with(Path("nextflow_schema.json"))

    @mock.patch("nf_core.pipelines.schema.PipelineSchema.get_schema_path")
    def test_schema_lint_filename(self, mock_get_schema_path):
        """Test nf-core pipelines schema lint accepts a filename"""
        cmd = ["pipelines", "schema", "lint", "some_other_filename"]
        with self.runner.isolated_filesystem():
            with open("some_other_filename", "w") as f:
                f.write("{}")
            self.invoke_cli(cmd)
            mock_get_schema_path.assert_called_with(Path("some_other_filename"))

    @mock.patch("nf_core.pipelines.create_logo.create_logo")
    def test_create_logo(self, mock_create_logo):
        # Set up the mock to return a specific value

        cmd = ["pipelines", "create-logo", "test"]
        result = self.invoke_cli(cmd)

        mock_create_logo.assert_called_with("test", Path.cwd(), None, "light", 2300, "png", False)
        assert result.exit_code == 0



import logging
import sys

import rich

from nf_core.utils import rich_force_colors

log = logging.getLogger(__name__)
stdout = rich.console.Console(force_terminal=rich_force_colors())


def modules_list_remote(ctx, keywords, json):
    """
    List modules in a remote GitHub repo [dim i](e.g [link=https://github.com/nf-core/modules]nf-core/modules[/])[/].
    """
    from nf_core.modules.list import ModuleList

    try:
        module_list = ModuleList(
            ".",
            True,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        stdout.print(module_list.list_components(keywords, json))
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)


def modules_list_local(ctx, keywords, json, directory):  # pylint: disable=redefined-builtin
    """
    List modules installed locally in a pipeline
    """
    from nf_core.modules.list import ModuleList

    try:
        module_list = ModuleList(
            directory,
            False,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        stdout.print(module_list.list_components(keywords, json))
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


def modules_install(ctx, tool, directory, prompt, force, sha):
    """
    Install DSL2 modules within a pipeline.

    Fetches and installs module files from a remote repo e.g. nf-core/modules.
    """
    from nf_core.modules.install import ModuleInstall

    try:
        module_install = ModuleInstall(
            directory,
            force,
            prompt,
            sha,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        exit_status = module_install.install(tool)
        if not exit_status:
            sys.exit(1)
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


def modules_update(
    ctx,
    tool,
    directory,
    force,
    prompt,
    sha,
    install_all,
    preview,
    save_diff,
    update_deps,
    limit_output,
):
    """
    Update DSL2 modules within a pipeline.

    Fetches and updates module files from a remote repo e.g. nf-core/modules.
    """
    from nf_core.modules.update import ModuleUpdate

    try:
        module_install = ModuleUpdate(
            directory,
            force,
            prompt,
            sha,
            install_all,
            preview,
            save_diff,
            update_deps,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
            limit_output,
        )
        exit_status = module_install.update(tool)
        if not exit_status and install_all:
            sys.exit(1)
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


def modules_patch(ctx, tool, directory, remove):
    """
    Create a patch file for minor changes in a module

    Checks if a module has been modified locally and creates a patch file
    describing how the module has changed from the remote version
    """
    from nf_core.modules.patch import ModulePatch

    try:
        module_patch = ModulePatch(
            directory,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        if remove:
            module_patch.remove(tool)
        else:
            module_patch.patch(tool)
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


def modules_remove(ctx, directory, tool):
    """
    Remove a module from a pipeline.
    """
    from nf_core.modules.remove import ModuleRemove

    try:
        module_remove = ModuleRemove(
            directory,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        module_remove.remove(tool)
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)


def modules_create(
    ctx,
    tool,
    directory,
    author,
    label,
    meta,
    no_meta,
    force,
    conda_name,
    conda_package_version,
    empty_template,
    migrate_pytest,
):
    """
    Create a new DSL2 module from the nf-core template.

    If the specified directory is a pipeline, this function creates a file called
    'modules/local/tool_subtool.nf'

    If the specified directory is a clone of nf-core/modules, it creates or modifies files
    in 'modules/', 'tests/modules' and 'tests/config/pytest_modules.yml'
    """
    # Combine two bool flags into one variable
    has_meta = None
    if meta and no_meta:
        log.critical("Both arguments '--meta' and '--no-meta' given. Please pick one.")
    elif meta:
        has_meta = True
    elif no_meta:
        has_meta = False

    from nf_core.modules.create import ModuleCreate

    # Run function
    try:
        module_create = ModuleCreate(
            directory,
            tool,
            author,
            label,
            has_meta,
            force,
            conda_name,
            conda_package_version,
            empty_template,
            migrate_pytest,
        )
        module_create.create()
    except UserWarning as e:
        log.critical(e)
        sys.exit(1)
    except LookupError as e:
        log.error(e)
        sys.exit(1)


def modules_test(ctx, tool, directory, no_prompts, update, once, profile, migrate_pytest):
    """
    Run nf-test for a module.

    Given the name of a module, runs the nf-test command to test the module and generate snapshots.
    """
    from nf_core.components.components_test import ComponentsTest

    if migrate_pytest:
        modules_create(
            ctx,
            tool,
            directory,
            author="",
            label="",
            meta=True,
            no_meta=False,
            force=False,
            conda_name=None,
            conda_package_version=None,
            empty_template=False,
            migrate_pytest=migrate_pytest,
        )
    try:
        module_tester = ComponentsTest(
            component_type="modules",
            component_name=tool,
            directory=directory,
            no_prompts=no_prompts,
            update=update,
            once=once,
            remote_url=ctx.obj["modules_repo_url"],
            branch=ctx.obj["modules_repo_branch"],
            verbose=ctx.obj["verbose"],
            profile=profile,
        )
        module_tester.run()
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)


def modules_lint(ctx, tool, directory, registry, key, all, fail_warned, local, passed, sort_by, fix_version, fix):
    """
    Lint one or more modules in a directory.

    Checks DSL2 module code against nf-core guidelines to ensure
    that all modules follow the same standards.

    Test modules within a pipeline or a clone of the
    nf-core/modules repository.
    """
    from nf_core.components.lint import LintExceptionError
    from nf_core.modules.lint import ModuleLint

    try:
        module_lint = ModuleLint(
            directory,
            fail_warned=fail_warned,
            fix=fix,
            registry=ctx.params["registry"],
            remote_url=ctx.obj["modules_repo_url"],
            branch=ctx.obj["modules_repo_branch"],
            no_pull=ctx.obj["modules_repo_no_pull"],
            hide_progress=ctx.obj["hide_progress"],
        )
        module_lint.lint(
            module=tool,
            registry=registry,
            key=key,
            all_modules=all,
            print_results=True,
            local=local,
            show_passed=passed,
            sort_by=sort_by,
            fix_version=fix_version,
        )
        if len(module_lint.failed) > 0:
            sys.exit(1)
    except LintExceptionError as e:
        log.error(e)
        sys.exit(1)
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)


def modules_info(ctx, tool, directory):
    """
    Show developer usage information about a given module.

    Parses information from a module's [i]meta.yml[/] and renders help
    on the command line. A handy equivalent to searching the
    [link=https://nf-co.re/modules]nf-core website[/].

    If run from a pipeline and a local copy of the module is found, the command
    will print this usage info.
    If not, usage from the remote modules repo will be shown.
    """
    from nf_core.modules.info import ModuleInfo

    try:
        module_info = ModuleInfo(
            directory,
            tool,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        stdout.print(module_info.get_component_info())
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


def modules_bump_versions(ctx, tool, directory, all, show_all):
    """
    Bump versions for one or more modules in a clone of
    the nf-core/modules repo.
    """
    from nf_core.modules.bump_versions import ModuleVersionBumper
    from nf_core.modules.modules_utils import ModuleExceptionError

    try:
        version_bumper = ModuleVersionBumper(
            directory,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        version_bumper.bump_versions(module=tool, all_modules=all, show_uptodate=show_all)
    except ModuleExceptionError as e:
        log.error(e)
        sys.exit(1)
    except (UserWarning, LookupError) as e:
        log.critical(e)
        sys.exit(1)



from pathlib import Path

import yaml

import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintActionsSchemaValidation(TestLint):
    def setUp(self) -> None:
        super().setUp()
        self.new_pipeline = self._make_pipeline_copy()
        self.awstest_yml_path = Path(self.new_pipeline) / ".github" / "workflows" / "awstest.yml"
        with open(self.awstest_yml_path) as fh:
            self.awstest_yml = yaml.safe_load(fh)

    def test_actions_schema_validation_missing_jobs(self):
        """Missing 'jobs' field should result in failure"""

        self.awstest_yml.pop("jobs")
        with open(self.awstest_yml_path, "w") as fh:
            yaml.dump(self.awstest_yml, fh)

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()

        results = lint_obj.actions_schema_validation()

        assert "Workflow validation failed for awstest.yml: 'jobs' is a required property" in results["failed"][0]

    def test_actions_schema_validation_missing_on(self):
        """Missing 'on' field should result in failure"""

        self.awstest_yml.pop(True)
        with open(self.awstest_yml_path, "w") as fh:
            yaml.dump(self.awstest_yml, fh)

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()

        results = lint_obj.actions_schema_validation()

        assert results["failed"][0] == "Missing 'on' keyword in awstest.yml"
        assert "Workflow validation failed for awstest.yml: 'on' is a required property" in results["failed"][1]

    def test_actions_schema_validation_fails_for_additional_property(self):
        """Missing 'jobs' field should result in failure"""

        self.awstest_yml["not_jobs"] = self.awstest_yml["jobs"]
        with open(self.awstest_yml_path, "w") as fh:
            yaml.dump(self.awstest_yml, fh)

        lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)
        lint_obj._load()

        results = lint_obj.actions_schema_validation()

        assert (
            "Workflow validation failed for awstest.yml: Additional properties are not allowed ('not_jobs' was unexpected)"
            in results["failed"][0]
        )



import logging
import os

log = logging.getLogger(__name__)


def module_deprecations(_, module):
    """
    Check that the modules are up to the latest nf-core standard
    """
    module.wf_path = module.component_dir
    if "functions.nf" in os.listdir(module.component_dir):
        module.failed.append(
            (
                "module_deprecations",
                "Deprecated file `functions.nf` found. No longer required for the latest nf-core/modules syntax!",
                module.component_dir,
            )
        )



from pathlib import Path

import yaml


def actions_ci(self):
    """Checks that the GitHub Actions pipeline CI (Continuous Integration) workflow is valid.

    The ``.github/workflows/ci.yml`` GitHub Actions workflow runs the pipeline on a minimal test
    dataset using ``-profile test`` to check that no breaking changes have been introduced.
    Final result files are not checked, just that the pipeline exists successfully.

    This lint test checks this GitHub Actions workflow file for the following:

    * Workflow must be triggered on the following events:

      .. code-block:: yaml

         on:
             push:
             branches:
                 - dev
             pull_request:
             release:
             types: [published]

    * The minimum Nextflow version specified in the pipeline's ``nextflow.config`` matches that defined by ``NXF_VER`` in the test matrix:

      .. code-block:: yaml
         :emphasize-lines: 4

         strategy:
           matrix:
             # Nextflow versions: check pipeline minimum and current latest
             NXF_VER: ['19.10.0', '']

      .. note:: These ``matrix`` variables run the test workflow twice, varying the ``NXF_VER`` variable each time.
                This is used in the ``nextflow run`` commands to test the pipeline with both the latest available version
                of the pipeline (``''``) and the stated minimum required version.
    """
    passed = []
    failed = []
    fn = Path(self.wf_path, ".github", "workflows", "ci.yml")

    # Return an ignored status if we can't find the file
    if not fn.is_file():
        return {"ignored": ["'.github/workflows/ci.yml' not found"]}

    try:
        with open(fn) as fh:
            ciwf = yaml.safe_load(fh)
    except Exception as e:
        return {"failed": [f"Could not parse yaml file: {fn}, {e}"]}

    # Check that the action is turned on for the correct events
    try:
        # NB: YAML dict key 'on' is evaluated to a Python dict key True
        if "dev" not in ciwf[True]["push"]["branches"]:
            raise AssertionError()
        pr_subtree = ciwf[True]["pull_request"]
        if not (
            pr_subtree is None
            or ("branches" in pr_subtree and "dev" in pr_subtree["branches"])
            or ("ignore_branches" in pr_subtree and "dev" not in pr_subtree["ignore_branches"])
        ):
            raise AssertionError()
        if "published" not in ciwf[True]["release"]["types"]:
            raise AssertionError()
    except (AssertionError, KeyError, TypeError):
        failed.append("'.github/workflows/ci.yml' is not triggered on expected events")
    else:
        passed.append("'.github/workflows/ci.yml' is triggered on expected events")

    # Check that we are testing the minimum nextflow version
    try:
        nxf_ver = ciwf["jobs"]["test"]["strategy"]["matrix"]["NXF_VER"]
        if not any(i == self.minNextflowVersion for i in nxf_ver):
            raise AssertionError()
    except (KeyError, TypeError):
        failed.append("'.github/workflows/ci.yml' does not check minimum NF version")
    except AssertionError:
        failed.append(
            f"Minimum pipeline NF version '{self.minNextflowVersion}' is not tested in '.github/workflows/ci.yml'"
        )
    else:
        passed.append("'.github/workflows/ci.yml' checks minimum NF version")

    return {"passed": passed, "failed": failed}
